
from typing import List

from dataclasses import dataclass


@dataclass
class OglDimensions:

    width:  int = 0
    height: int = 0

    @classmethod
    def deSerialize(cls, value: str) -> 'OglDimensions':

        dimensions: OglDimensions = OglDimensions()

        widthHeight: List[str] = value.split(sep=',')

        assert len(widthHeight) == 2, 'Incorrectly formatted dimensions'
        assert value.replace(',', '', 1).isdigit(), 'String must be numeric'

        dimensions.width  = int(widthHeight[0])
        dimensions.height = int(widthHeight[1])

        return dimensions

    def __str__(self):
        return f'{self.width},{self.height}'

    def __repr__(self):
        return self.__str__()




from typing import Union

from logging import Logger
from logging import getLogger

from miniogl.Shape import Shape
from miniogl.SizerShape import SizerShape


class Diagram:

    clsLogger: Logger = getLogger(__name__)
    """
    A diagram contains shapes and is manages them.
    
    It knows every shapes that can be clicked, selected, and moved.
    """
    def __init__(self, panel):
        """
        Constructor.

        @param  panel : the panel on which to draw
        """
        self._panel = panel
        self._shapes = []        # all selectable shapes
        self._parentShapes = []  # all first level shapes

    def AddShape(self, shape, withModelUpdate: bool = True):
        """
        Add a shape to the diagram.
        This is the correct way to do it. Don't use Shape.Attach(diagram)!

        Args:
            shape:  the shape to add
            withModelUpdate:
        """
        if shape not in self._shapes:
            self._shapes.append(shape)
        if shape not in self._parentShapes and shape.GetParent() is None:
            self._parentShapes.append(shape)

        self.clsLogger.debug(f'.AddShape before shape.Attach()=> {shape} withModelUpdate {withModelUpdate}')
        shape.Attach(self)

        # makes the shape's model (MVC pattern) have the right values depending on
        # the diagram frame state.
        if withModelUpdate:
            shape.UpdateModel()

    def DeleteAllShapes(self):
        """
        Delete all shapes in the diagram.
        """
        while self._shapes:
            self._shapes[0].Detach()
        self._shapes = []
        self._parentShapes = []

    def RemoveShape(self, shape: Union[Shape, SizerShape]):
        """
        Remove a shape from the diagram. Use Shape.Detach() instead!
        This also works, but it not the better way.

        TODO:  Use a Union for now;  I think the correct type is just Shape.  But,
        I need to see how Pyut uses it

        Args:
            shape:

        """
        Diagram.clsLogger.debug(f'Determine what got passed in: {shape=}')
        if isinstance(shape, SizerShape):
            self.clsLogger.warning(f'Removing SizerShape')
        if shape in self._shapes:
            self._shapes.remove(shape)
        if shape in self._parentShapes:
            self._parentShapes.remove(shape)

    def GetShapes(self):
        """
        Return a list of the shapes in the diagram.
        It is a copy of the original. You cannot detach or add shapes to the
        diagram this way.

        @return Shape []
        """
        return self._shapes[:]

    def GetParentShapes(self):
        """
        Return a list of the parent shapes in the diagram.
        It is a copy of the original. You cannot detach or add shapes to the
        diagram this way.

        @return Shape []
        """
        return self._parentShapes[:]

    def GetPanel(self):
        """
        Return the panel associated with this diagram.

        @return DiagramFrame
        """
        return self._panel

    def MoveToFront(self, shape: Shape):
        """
        Move the given shape to the end of the display list => last drawn.

        Args:
            shape: The shape to move
        """
        shapes = [shape] + shape.GetAllChildren()
        for s in shapes:
            self._shapes.remove(s)
        self._shapes = self._shapes + shapes

    def MoveToBack(self, shape: Shape):
        """
        Move the given shape to the start of the display list => first drawn.

        Args:
            shape: The shape to move
        """
        shapes = [shape] + shape.GetAllChildren()
        for s in shapes:
            self._shapes.remove(s)
        self._shapes = shapes + self._shapes




from logging import Logger
from logging import getLogger

from wx import DC
from wx import Pen

from wx import PENSTYLE_SHORT_DASH
from wx import PENSTYLE_LONG_DASH
from wx import RED_PEN
from wx import BLACK_PEN
from wx import WHITE_BRUSH

from pyutmodel.PyutLink import PyutLink

from ogl.OglLink import OglLink
from ogl.OglClass import OglClass

# Kind of labels
[CENTER] = list(range(1))


class OglInterface(OglLink):

    clsLogger: Logger = getLogger(__name__)

    """
    Graphical OGL representation of an interface link.
    This class provide the methods for drawing an interface link between
    two classes of an UML diagram. Add labels to an OglLink.
    """
    def __init__(self, srcShape: OglClass, pyutLink: PyutLink, dstShape: OglClass, srcPos=None, dstPos=None):

        """

        Args:
            srcShape:  Source shape
            pyutLink:  Conceptual links associated with the graphical links.
            dstShape:  Destination shape
            srcPos:    Position of source      Override location of input source
            dstPos:    Position of destination Override location of input destination

        """
        super().__init__(srcShape, pyutLink, dstShape, srcPos=srcPos, dstPos=dstPos)

        self.SetPen(Pen("BLACK", 1, PENSTYLE_LONG_DASH))
        self.SetBrush(WHITE_BRUSH)
        self._labels = {CENTER: self.AddText(0, 0, "")}

        # Initialize labels objects
        self.updateLabels()
        self.SetDrawArrow(True)

    def updateLabels(self):
        """
        Update the labels according to the link.
        """
        def prepareLabel(textShape, text):
            # If label should be drawn
            if text.strip() != "":
                textShape.SetText(text)
                # textShape.Show(True)
                textShape.SetVisible(True)
            else:
                # textShape.Show(False)
                textShape.SetVisible(False)

        # Prepares labels
        prepareLabel(self._labels[CENTER], self._link.name)

    def getLabels(self):
        """
        Get the labels.

        @return TextShape []
        @since 1.0
        """
        return self._labels

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Called for drawing of interface links.
        OglLink drew regular lines
        I need dashed lines for an interface

        Args:
            dc: Device context
            withChildren:   Draw the children or not

        """
        self.updateLabels()
        if self._visible:
            line = self.GetSegments()
            if self._selected:
                dc.SetPen(RED_PEN)

            if self._spline:
                dc.DrawSpline(line)
            else:
                pen: Pen = dc.GetPen()              #
                pen.SetStyle(PENSTYLE_SHORT_DASH)   # This is what is different from OglLink.Draw(..)
                dc.SetPen(pen)                      #
                dc.DrawLines(line)

            for control in self._controls:
                control.Draw(dc)

            if self._selected:
                self._srcAnchor.Draw(dc)
                self._dstAnchor.Draw(dc)
            dc.SetPen(BLACK_PEN)

            if self._drawArrow:
                u, v = line[-2], line[-1]
                self.DrawArrow(dc, u, v)

            if withChildren is True:
                self.DrawChildren(dc)

    def __repr__(self):
        return f'OglInterface - {super().__repr__()}'




from miniogl.LinePoint import LinePoint


class ControlPoint(LinePoint):
    """
    This is a point which guides lines.
    A control point must be linked to a line, it has no life by itself.

    If you remove the last line of a control point, the control point will
    automatically be erased.
    """
    def __init__(self, x: int, y: int, parent=None):
        """

        Args:
            x:  abscissa of the point
            y:  ordinate of the point
            parent:     parent Shape
        """
        super().__init__(x, y, parent)
        self.SetVisible(False)

    def RemoveLine(self, line):
        """
        Remove a line from the point.
        If there are no more lines for this point, it is automatically
        detached.

        Args:
            line:   The line to remove
        """
        super(ControlPoint, self).RemoveLine(line)
        if len(self._lines) == 0:
            self.Detach()

    def __eq__(self, other) -> bool:

        if isinstance(other, ControlPoint):
            # noinspection PyProtectedMember
            if self._x == other._x and self._y == other._y and self._id == other._id:
                return True
            else:
                return False
        else:
            return False

    def __repr__(self):
        return f'ControlPoint@ {self._x},{self._y} {self._id=} {self._visible=}'




from logging import Logger
from logging import getLogger

from wx import BLACK
from wx import WHITE
from wx import PENSTYLE_SOLID

from wx import Font
from wx import Colour

from wx import DC
from wx import MemoryDC

from miniogl.Shape import Shape
from miniogl.RectangleShape import RectangleShape
from miniogl.TextShapeModel import TextShapeModel


class TextShape(RectangleShape):

    clsLogger: Logger = getLogger(__name__)
    """
    A text shape that can be attached to another shape standalone).
    """
    def __init__(self, x: int, y: int, text: str, parent=None, font: Font = None):
        """

        Args:
            x:          x position of the point
            y:          y position of the point
            text:       the text that the shape displays
            parent:     parent shape
            font:       Font to use
        """
        self._text:  str    = ''

        super().__init__(x, y, 0, 0, parent)

        self._color: Colour = BLACK
        self.SetText(text)

        self._drawFrame: bool = False
        self._resizable: bool = False
        self._textBack:  Colour = WHITE    # text background colour

        self._model: TextShapeModel = TextShapeModel(self)
        self._font:  Font = font

    def Attach(self, diagram):
        """
        Do not use this method, use Diagram.AddShape instead !!!
        Attach the shape to a diagram.
        When you create a new shape, you must attach it to a diagram before
        you can see it. This method is used internally by Diagram.AddShape.

        Args:
            diagram
        """
        # RectangleShape.Attach(self, diagram)
        super().Attach(diagram)
        self._textBack = self._diagram.GetPanel().GetBackgroundColour()

    def GetText(self) -> str:
        """
        Returns:  The text that the text shape displays
        """
        return self._text

    def SetText(self, text: str):
        """
        Set the text that the shape displays

        Args:
              text
        """
        self._text = text
        self._width, self._height = MemoryDC().GetTextExtent(text)

    def SetTextBackground(self, color: Colour):
        """
        Set the text background color.

        Args:
             color
        """
        self._textBack = color

    def GetTextBackground(self) -> Colour:
        """
        Get the text background color.

        Returns:
             the text background color
        """
        return self._textBack

    def Draw(self, dc: DC, withChildren: bool = True):
        """
        Draw the text on the dc.

        Args:
            dc
            withChildren
        """
        if self._visible:
            RectangleShape.Draw(self, dc, False)
            dc.SetTextForeground(self._color)
            dc.SetBackgroundMode(PENSTYLE_SOLID)
            dc.SetTextBackground(self._textBack)
            x, y = self.GetPosition()

            # to draw the text shape with its own font size
            saveFont: Font = dc.GetFont()
            if self.GetFont() is not None:
                dc.SetFont(self.GetFont())

            dc.DrawText(self._text, x, y)
            dc.SetFont(saveFont)

            if withChildren:
                self.DrawChildren(dc)

    def DrawBorder(self, dc: DC):
        """
        Draw the border of the shape, for fast rendering.

        Args:
            dc
        """
        if self._selected:
            RectangleShape.DrawBorder(self, dc)
        else:
            Shape.DrawBorder(self, dc)

    def GetColor(self) -> Colour:
        """
        Return the text color

        Returns wx.Colour
        """
        return self._color

    def SetColor(self, color: Colour):
        """
        Set the color of the text.

        Args:
             color
        """
        self._color = color

    def UpdateFromModel(self):
        """
        Updates the shape position and size from the model in light of a
        change of state of the diagram frame.  Here it is only for the zoom
        """

        # change the position and size of the shape from the model
        # RectangleShape.UpdateFromModel(self)
        super().UpdateFromModel()
        # get the diagram frame ratio between the shape and the model
        ratio = self.GetDiagram().GetPanel().GetCurrentZoom()

        fontSize = round(self.GetModel().GetFontSize() * ratio)
        TextShape.clsLogger.debug(f'UpdateFromModel - ratio: {ratio}')

        # set the new font size
        if self._font is not None:
            self._font.SetPointSize(fontSize)

    def UpdateModel(self):
        """
        Updates the model when the shape (view) is displaced or resized.
        """
        # change the coordinates and size of model
        # RectangleShape.UpdateModel(self)
        super().UpdateModel()

        # get the ratio between the model and the shape (view) from
        # the diagram frame where the shape is displayed.
        ratio = self.GetDiagram().GetPanel().GetCurrentZoom()

        # TextShape.clsLogger.debug(f'UpdateModel - ratio: {ratio}')
        if self.GetFont() is not None:
            fontSize = self.GetFont().GetPointSize() // ratio
            self.GetModel().SetFontSize(fontSize)

    def GetFont(self) -> Font:
        """

        Returns:  The font used by the text shape

        """
        return self._font

    def __repr__(self):
        x, y = self.GetPosition()
        return f'TextShape[{self._text=} position: ({x},{y}])'




from typing import cast

from dataclasses import dataclass

from wx import Point

from miniogl.Shape import Shape


@dataclass
class ShapeSelectedEventData:

    shape:    Shape = cast(Shape, None)
    position: Point = cast(Point, None)




from typing import cast

from logging import Logger
from logging import getLogger

from wx import Brush
from wx import Colour
from wx import DC

from ogl.OglObject import OglObject

from ogl.preferences.OglPreferences import OglPreferences

from pyutmodel.PyutNote import PyutNote

from ogl.OglUtils import OglUtils


class OglNote(OglObject):
    """
    This is an OGL object that represents a UML note in diagrams.
    A note may be linked
    with all links except Inheritance and Interface.

    For more instructions about how to create an OGL object, please refer
    to the `OglObject` class.
    """

    MARGIN: int = 10

    def __init__(self, pyutNote=None, w: int = 0, h: int = 0):
        """

        Args:
            pyutNote:   A PyutNote Object
            w:          Default width override
            h:          Default height override
        """

        # Init pyutObject (coming from OglObject)
        if pyutNote is None:
            pyutObject = PyutNote()
        else:
            pyutObject = pyutNote

        width:  int = w
        height: int = h
        prefs: OglPreferences = OglPreferences()
        if width == 0:
            width = prefs.noteDimensions.width
        if height == 0:
            height = prefs.noteDimensions.height

        super().__init__(pyutObject, width=width, height=height)

        self.logger: Logger = getLogger(__name__)
        self.SetBrush(Brush(Colour(255, 255, 230)))

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Paint handler, draws the content of the shape.

        Args:
            dc:     device context to draw to
            withChildren:   Redraw children or not
        """
        super().Draw(dc)
        dc.SetFont(self._defaultFont)

        w, h = self.GetSize()

        try:
            noteContent = cast(PyutNote, self.pyutObject).content
            lines = OglUtils.lineSplitter(noteContent, dc, w - 2 * OglNote.MARGIN)
        except (ValueError, Exception) as e:
            self.logger.error(f"Unable to display note - {e}")
            return

        baseX, baseY = self.GetPosition()

        dc.SetClippingRegion(baseX, baseY, w, h)

        x = baseX + OglNote.MARGIN
        y = baseY + OglNote.MARGIN

        for line in range(len(lines)):
            dc.DrawText(lines[line], x, y + line * (dc.GetCharHeight() + 5))

        dc.DrawLine(baseX + w - OglNote.MARGIN, baseY, baseX + w, baseY + OglNote.MARGIN)

        dc.DestroyClippingRegion()

    def __repr__(self):
        pyutNote: PyutNote = cast(PyutNote, self.pyutObject)
        if pyutNote is None:
            return f'Anonymous Note'
        else:
            return f'{pyutNote.content}'




from enum import Enum

from wx.lib.newevent import NewEvent

#
# Constructor returns a tuple; First is the event,  The second is the binder
#
ShapeSelectedEvent,        EVT_SHAPE_SELECTED   = NewEvent()
CutOglClassEvent,          EVT_CUT_OGL_CLASS    = NewEvent()
DiagramFrameModifiedEvent, EVT_DIAGRAM_FRAME_MODIFIED = NewEvent()

RequestLollipopLocationEvent, EVT_REQUEST_LOLLIPOP_LOCATION = NewEvent()
CreateLollipopInterfaceEvent, EVT_CREATE_LOLLIPOP_INTERFACE = NewEvent()


class OglEventType(Enum):
    """

    """

    ShapeSelected           = 'ShapeSelected'
    CutOglClass             = 'CutOglClass'
    DiagramFrameModified    = 'DiagramFrameModified'
    RequestLollipopLocation = 'RequestLollipopLocation'
    CreateLollipopInterface = 'CreateLollipopInterface'




from typing import cast

from logging import Logger
from logging import getLogger

from wx import BLACK_PEN
from wx import Colour
from wx import ColourDatabase
from wx import EVT_MENU
from wx import FONTSTYLE_ITALIC
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_BOLD
from wx import FONTWEIGHT_NORMAL

from wx import Bitmap
from wx import CommandEvent
from wx import DC
from wx import Font
from wx import Menu
from wx import MenuItem
from wx import MouseEvent

from wx import RED_PEN


from miniogl.DiagramFrame import DiagramFrame

from ogl.OglObject import OglObject
from ogl.OglTextFontFamily import OglTextFontFamily

from ogl.OglUtils import OglUtils

from ogl.preferences.OglPreferences import OglPreferences

from ogl.resources.img.textdetails.DecreaseTextSize import embeddedImage as DecreaseTextSize
from ogl.resources.img.textdetails.IncreaseTextSize import embeddedImage as IncreaseTextSize

from pyutmodel.PyutText import PyutText
[
    ID_MENU_INCREASE_SIZE,
    ID_MENU_DECREASE_SIZE,
    ID_MENU_BOLD_TEXT,
    ID_MENU_ITALIC_TEXT
]  = OglUtils.assignID(4)

TEXT_SIZE_INCREMENT: int = 2
TEXT_SIZE_DECREMENT: int = 2


class OglText(OglObject):
    """
    Draws resizeable boxes of text with no visible boundaries
    PyutText should not have graphical attributes.
    """
    MARGIN: int = 5

    def __init__(self, pyutText: PyutText, width: int = 0, height: int = 0):    # TODO make default text size a preference
        """
        Args:
            pyutText:   Associated PyutText instance
            width:      Initial width
            height:     Initial height
        """
        w: int = width
        h: int = height

        # Use preferences to get initial size if not specified
        preferences: OglPreferences = OglPreferences()

        if width == 0:
            w = preferences.textDimensions.width
        if height == 0:
            h = preferences.textDimensions.height

        super().__init__(pyutObject=pyutText, width=w, height=h)

        self.logger: Logger = getLogger(__name__)

        self._drawFrame:      bool = False
        self._textFontFamily: OglTextFontFamily = preferences.textFontFamily
        self._textSize:       int  = preferences.textFontSize
        self._isBold:         bool = preferences.textBold
        self._isItalicized:   bool = preferences.textItalicize

        self._textFont:  Font            = self._defaultFont.GetBaseFont()
        self._textFont.SetFamily(OglUtils.oglFontFamilyToWxFontFamily(self._textFontFamily))

        self._redColor:   Colour  = ColourDatabase().Find('Red')
        self._blackColor: Colour = ColourDatabase().Find('Black')

        self.__initializeTextDisplay()
        self._menu: Menu = cast(Menu, None)

    @property
    def pyutText(self) -> PyutText:
        return self._pyutObject

    @pyutText.setter
    def pyutText(self, newValue: PyutText):
        self._pyutObject = newValue

    @property
    def textSize(self) -> int:
        return self._textSize

    @textSize.setter
    def textSize(self, newSize: int):
        self._textSize = newSize

    @property
    def isBold(self) -> bool:
        return self._isBold

    @isBold.setter
    def isBold(self, newValue: bool):
        self._isBold = newValue

    @property
    def isItalicized(self) -> bool:
        return self._isItalicized

    @isItalicized.setter
    def isItalicized(self, newValue: bool):
        self._isItalicized = newValue

    @property
    def textFontFamily(self) -> OglTextFontFamily:
        return self._textFontFamily

    @textFontFamily.setter
    def textFontFamily(self, newValue: OglTextFontFamily):
        self._textFontFamily = newValue

    def OnLeftUp(self, event: MouseEvent):
        """
        Implement this method so we can snap Ogl objects

        Args:
            event:  the mouse event
        """
        super().OnLeftUp(event)

    def OnRightDown(self, event: MouseEvent):
        """
        Callback for right clicks
        """
        if self._menu is None:
            self._menu = self._createMenu()

        frame = self._diagram.GetPanel()

        x: int = event.GetX()
        y: int = event.GetY()
        self.logger.debug(f'OglClass - x,y: {x},{y}')

        frame.PopupMenu(self._menu, x, y)

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Paint handler, draws the content of the shape.

        Args:
            dc:     device context to draw to
            withChildren:   Redraw children or not
        """
        if self._selected:
            dc.SetPen(RED_PEN)
            dc.SetTextForeground(self._redColor)
        else:
            dc.SetPen(BLACK_PEN)
            dc.SetTextForeground(self._blackColor)

        OglObject.Draw(self, dc)
        dc.SetFont(self._textFont)

        w, h = self.GetSize()

        baseX, baseY = self.GetPosition()

        dc.SetClippingRegion(baseX, baseY, w, h)

        noteContent = self.pyutObject.content
        lines = OglUtils.lineSplitter(noteContent, dc, w - 2 * OglText.MARGIN)

        x = baseX + OglText.MARGIN
        y = baseY + OglText.MARGIN

        for line in range(len(lines)):
            dc.DrawText(lines[line], x, y + line * (dc.GetCharHeight() + 5))

        dc.DestroyClippingRegion()

    def _createMenu(self) -> Menu:

        menu: Menu = Menu()

        increaseItem: MenuItem = menu.Append(ID_MENU_INCREASE_SIZE, 'Increase Size', 'Increase Text Size by 2 points')
        decreaseItem: MenuItem = menu.Append(ID_MENU_DECREASE_SIZE, 'Decrease Size', 'Decrease Text Size by 2 points')

        incBmp: Bitmap = IncreaseTextSize.GetBitmap()
        increaseItem.SetBitmap(incBmp)
        decBmp: Bitmap = DecreaseTextSize.GetBitmap()
        decreaseItem.SetBitmap(decBmp)

        boldItem:       MenuItem = menu.AppendCheckItem(ID_MENU_BOLD_TEXT,   item='Bold Text',      help='Set text to bold')
        italicizedItem: MenuItem = menu.AppendCheckItem(ID_MENU_ITALIC_TEXT, item='Italicize Text', help='Set text to italics')

        if self.isBold is True:
            boldItem.Check(check=True)
        if self.isItalicized is True:
            italicizedItem.Check(check=True)

        menu.Bind(EVT_MENU, self._onChangeTextSize, id=ID_MENU_INCREASE_SIZE)
        menu.Bind(EVT_MENU, self._onChangeTextSize, id=ID_MENU_DECREASE_SIZE)
        menu.Bind(EVT_MENU, self._onToggleBold,     id=ID_MENU_BOLD_TEXT)
        menu.Bind(EVT_MENU, self._onToggleItalicize, id=ID_MENU_ITALIC_TEXT)

        return menu

    def _onChangeTextSize(self, event: CommandEvent):
        """
        Callback for popup menu on class

        Args:
            event:
        """
        eventId:  int      = event.GetId()

        if eventId == ID_MENU_INCREASE_SIZE:
            self.textSize += TEXT_SIZE_INCREMENT
        elif eventId == ID_MENU_DECREASE_SIZE:
            self.textSize -= TEXT_SIZE_DECREMENT
        else:
            assert False, f'Unhandled text size event: {eventId}'

        self._textFont.SetPointSize(self.textSize)
        self.__updateDisplay()

    # noinspection PyUnusedLocal
    def _onToggleBold(self, event: CommandEvent):

        if self.isBold is True:
            self.isBold = False
            self._textFont.SetWeight(FONTWEIGHT_NORMAL)
        else:
            self.isBold = True
            self._textFont.SetWeight(FONTWEIGHT_BOLD)

        self.__updateDisplay()

    # noinspection PyUnusedLocal
    def _onToggleItalicize(self, event: CommandEvent):

        if self.isItalicized is True:
            self.isItalicized = False
            self._textFont.SetStyle(FONTSTYLE_NORMAL)
        else:
            self.isItalicized = True
            self._textFont.SetStyle(FONTSTYLE_ITALIC)

        self.__updateDisplay()

    def __updateDisplay(self):

        self.autoResize()

        frame: DiagramFrame = self._diagram.GetPanel()
        frame.Refresh()

    def __initializeTextDisplay(self):
        """
        Use the model to get other text attributes; We'll
        get what was specified or defaults
        """

        self._textFont.SetPointSize(self.textSize)

        if self.isBold is True:
            self._textFont.SetWeight(FONTWEIGHT_BOLD)
        if self.isItalicized is True:
            self._textFont.SetWeight(FONTWEIGHT_NORMAL)

        if self.isItalicized is True:
            self._textFont.SetStyle(FONTSTYLE_ITALIC)
        else:
            self._textFont.SetStyle(FONTSTYLE_NORMAL)

        self._textFont.SetPointSize(self.textSize)
        self._textFont.SetFamily(OglUtils.oglFontFamilyToWxFontFamily(self.textFontFamily))

    def __repr__(self):

        strMe: str = f"[OglText - name: '{self._pyutObject.name}' id: '{self._pyutObject.id}']"
        return strMe



# ----------------------------------------------------------------------
# This file was generated by /Users/humberto.a.sanchez.ii/PycharmProjects/PyUt/venv-pyenv-3.8.5-alternate/bin/img2py
#
from wx.lib.embeddedimage import PyEmbeddedImage

embeddedImage: PyEmbeddedImage = PyEmbeddedImage(
    b'iVBORw0KGgoAAAANSUhEUgAAABAAAAAQCAYAAAAf8/9hAAAAKnRFWHRDcmVhdGlvbiBUaW1l'
    b'AE1vIDMgTXJ6IDIwMDMgMjI6NDI6MTAgKzAxMDD/JlV7AAAAB3RJTUUH1AcHCBwd4e5wDAAA'
    b'AAlwSFlzAAAK8AAACvABQqw0mAAAAARnQU1BAACxjwv8YQUAAAIdSURBVHjatZPbTxNBFId/'
    b'O9t2S7EINKDB2oRgxYhakkZb45P+B8YXX7UJr6Y8+MS/gRohaqImokaMVzCEy4PxhlUf1JQg'
    b'XaCRSy0tXbZlL916drOikkbxwTP5kpk58+1szswA/x6czcZga9ELP2dg0Fl2COoL/RwG8IZm'
    b'ZX5L8g3s5FVu9PzRrkNCnbwjJWeOVCbwkjIrjr/KN9GMFYx0R2Ptw+lBVBgHxpjPgBGh7Dwz'
    b'17Ql6kOBB9tOYhdqafjzr67Cx2XxrDsS6xj5+hA6qyCZzJf0K8YQZV0EY22P67rCxfDbTs/+'
    b'e0IPLtOk30peQwMKGI5HY6GxhUfQUMb0VGG9dEG9izlkaM07Is+3n259EvB6vVllHi6P6+By'
    b'sNiCJSRxALfi0TOHxxefQuM0pKYlpdhL8owljxIJYpVXTpSOcy4lCI5DDXPC4WUduVblVPzY'
    b'2b0TmSGo1MQZSZUvqnfwZUOeNAtIGHwxrT2f9a1GGhtrdtM34HQwhAP7aj/kJqFUNMymJE2+'
    b'pN7GlCWPEebxZYkyrIItYh0pJBaa10INPrcfVNZMMQedNzAnSrrcT/JnSzB3fk18+yHDrngZ'
    b'OWoi3i81yZ3bm9wtvMCQFgv6Wp86gI+WMF5Nxi9HVkaedhGRyAqlPfm84ild1++TnLXlV0Rm'
    b's7z5Kpt9NxEkzEtST3yyq71cTa72FsyxYMtOQrKpKv8pfntx/zW+A1iI1PZ4wtPwAAAAAElF'
    b'TkSuQmCC')





from typing import Tuple
from typing import cast

from logging import Logger
from logging import getLogger

from math import sqrt

from deprecated import deprecated

from wx import BLACK_PEN
from wx import EVT_MENU

from wx import CommandEvent
from wx import Menu
from wx import MenuItem
from wx import MouseEvent


from pyutmodel.PyutLink import PyutLink
from pyutmodel.PyutLink import PyutLinks

from miniogl.AnchorPoint import AnchorPoint
from miniogl.ControlPoint import ControlPoint
from miniogl.LinePoint import LinePoint
from miniogl.LineShape import LineShape
from miniogl.Shape import Shape
from miniogl.ShapeEventHandler import ShapeEventHandler
from miniogl.AttachmentLocation import AttachmentLocation

from ogl.OglPosition import OglPosition

from ogl.IllegalOperationException import IllegalOperationException

from ogl.OglUtils import OglUtils

[
    MENU_ADD_BEND,
    MENU_REMOVE_BEND,
    MENU_TOGGLE_SPLINE,
]  = OglUtils.assignID(3)


class OglLink(LineShape, ShapeEventHandler):
    """
    Abstract class for graphical link.
    This class should be the base class for every type of link. It implements
    the following functions :
        - Link between objects position management
        - Control points (2)
        - Data layer link association
        - Source and destination objects

    You can inherit from this class to implement your favorite type of links
    like `OglAssociation`.

    There is a link factory (See `OglLinkFactory`) you can use to build
    the different type of links that exist.

    """
    clsLogger: Logger = getLogger(__name__)

    def __init__(self, srcShape, pyutLink, dstShape, srcPos=None, dstPos=None):
        """

        Args:
            srcShape:   Source shape
            pyutLink:   Conceptual links associated with the graphical links.
            dstShape:   Destination shape
            srcPos:     Position of source      Override location of input source
            dstPos:     Position of destination Override location of input destination
        """
        self._srcShape  = srcShape
        self._destShape = dstShape

        OglLink.clsLogger.debug(f'Input Override positions - srcPos: {srcPos} dstPos: {dstPos}')
        if srcPos is None and dstPos is None:
            srcX, srcY = self._srcShape.GetPosition()
            dstX, dstY = self._destShape.GetPosition()

            orient = OglLink.getOrient(srcX,  srcY, dstX, dstY)

            sw, sh = self._srcShape.GetSize()
            dw, dh = self._destShape.GetSize()
            if orient == AttachmentLocation.NORTH:
                srcX, srcY = sw//2, 0
                dstX, dstY = dw//2, dh
            elif orient == AttachmentLocation.SOUTH:
                srcX, srcY = sw//2, sh
                dstX, dstY = dw//2, 0
            elif orient == AttachmentLocation.EAST:
                srcX, srcY = sw, sh//2
                dstX, dstY = 0, dh//2
            elif orient == AttachmentLocation.WEST:
                srcX, srcY = 0, sh//2
                dstX, dstY = dw, dh//2

            # ============== Avoid over-lining; Added by C.Dutoit ================
            # lstAnchorsPoints = [anchor.GetRelativePosition() for anchor in srcSDInstance.GetAnchors()]
            # while (srcX, srcY) in lstAnchorsPoints:
            #     OglLink.clsLogger.warning(f'Over-lining in source shape')
            #     if orient == PyutAttachmentPoint.NORTH or orient == PyutAttachmentPoint.SOUTH:
            #         srcX += 10
            #     else:
            #         srcY += 10
            #
            # lstAnchorsPoints = [anchor.GetRelativePosition() for anchor in dstSDInstance.GetAnchors()]
            # while (dstX, dstY) in lstAnchorsPoints:
            #     from org.pyut.ogl.OglClass import OglClass
            #     dstSDInstance: OglClass = cast(OglClass, dstSDInstance)
            #     OglLink.clsLogger.warning(f'Over-lining in destination shape: {dstSDInstance.getPyutObject}')
            #     if orient == PyutAttachmentPoint.NORTH or orient == PyutAttachmentPoint.SOUTH:
            #         dstX += 10
            #     else:
            #         dstY += 10

            # =========== end avoid over-lining-Added by C.Dutoit ================
        else:
            # Use provided position
            (srcX, srcY) = srcPos
            (dstX, dstY) = dstPos

        srcAnchor: AnchorPoint = self._srcShape.AddAnchor(srcX, srcY)
        dstAnchor: AnchorPoint = self._destShape.AddAnchor(dstX, dstY)
        srcAnchor.SetPosition(srcX, srcY)
        dstAnchor.SetPosition(dstX, dstY)
        srcAnchor.SetVisible(False)
        dstAnchor.SetVisible(False)
        OglLink.clsLogger.debug(f'src anchor pos: {srcAnchor.GetPosition()} dst anchor pos {dstAnchor.GetPosition()}')
        srcAnchor.SetDraggable(True)
        dstAnchor.SetDraggable(True)
        # Init
        LineShape.__init__(self, srcAnchor, dstAnchor)
        # Set up painting colors
        self.SetPen(BLACK_PEN)
        # Keep reference to the PyutLink for mouse events, in order
        # to can find back the corresponding link
        if pyutLink is not None:
            self._link = pyutLink
        else:
            self._link = PyutLink()

    @staticmethod
    def getOrient(srcX, srcY, destX, destY) -> AttachmentLocation:
        """
        Giving a source and destination, returns where the destination
        is located according to the source.

        @param int srcX  : X pos of src point
        @param int srcY  : Y pos of src point
        @param int destX : X pos of dest point
        @param int destY : Y pos of dest point
        """
        deltaX = srcX - destX
        deltaY = srcY - destY
        if deltaX > 0:  # dest is not east
            if deltaX > abs(deltaY):  # dest is west
                return AttachmentLocation.WEST
            elif deltaY > 0:
                return AttachmentLocation.NORTH
            else:
                return AttachmentLocation.SOUTH
        else:  # dest is not west
            if -deltaX > abs(deltaY):  # dest is east
                return AttachmentLocation.EAST
            elif deltaY > 0:
                return AttachmentLocation.NORTH
            else:
                return AttachmentLocation.SOUTH

    @property
    def sourceShape(self):
        """
        Returns:  the source shape for this link.
        """
        return self._srcShape

    @property
    def destinationShape(self):
        """

        Returns:    The destination shape for this link.
        """
        return self._destShape

    @deprecated(reason='Use sourceShape property')
    def getSourceShape(self):
        """
        Returns:  the source shape for this link.
        """
        return self._srcShape

    @deprecated(reason='Use destinationShape property')
    def getDestinationShape(self):
        """

        Returns:    The destination shape for this link.
        """
        return self._destShape

    @deprecated('Use pyutObject property')
    def getPyutObject(self):
        """
        Returns the associated PyutLink.

        Returns: PyutLink

        """
        return self._link

    @deprecated('Use pyutObject property')
    def setPyutObject(self, pyutLink):
        """
        Sets the associated PyutLink.

        Args:
            pyutLink: link to associate
        """
        self._link = pyutLink

    @property
    def pyutObject(self) -> PyutLink:
        return self._link

    @pyutObject.setter
    def pyutObject(self, pyutLink: PyutLink):
        self._link = pyutLink

    def getAnchors(self) -> Tuple[AnchorPoint, AnchorPoint]:
        return self._srcAnchor, self._dstAnchor

    def Detach(self):
        """
        Detach the line and all its line points, including source and destination.
        """
        if self._diagram is not None and not self._protected:
            LineShape.Detach(self)
            self._srcAnchor.SetProtected(False)
            self._dstAnchor.SetProtected(False)
            self._srcAnchor.Detach()
            self._dstAnchor.Detach()
            self._detachFromOglEnds()
            self._detachModel()

    def _detachModel(self):
        """
        From the data model of the source remove the
        data model link
        """
        from typing import Union
        from pyutmodel.PyutClass import PyutClass
        from pyutmodel.PyutNote import PyutNote
        try:
            # self._link.getSource().links.remove(self._link)
            pyutSrc: Union[PyutClass, PyutNote] = self._link.getSource()
            links: PyutLinks = pyutSrc.links
            links.remove(self._link)
        except ValueError as ve:
            self.clsLogger.warning(f'Ignoring source removal error: {ve}')

    def _detachFromOglEnds(self):
        """
        Remove us (self) from the links list in each of the ends

        """
        # Do local imports because of these incestuous self references
        from typing import Union
        from typing import List
        from ogl.OglClass import OglClass
        from ogl.OglNote import OglNote
        try:
            # self.getSourceShape().links.remove(self)
            src: Union[OglClass, OglNote] = self.getSourceShape()
            links: List[OglLink] = src.links
            links.remove(self)
        except ValueError as ve:
            self.clsLogger.warning(f'Ignoring source removal error: {ve}')

        try:
            # self.getDestinationShape().links.remove(self)
            dest: Union[OglClass, OglNote] = self.getDestinationShape()
            links = dest.links
            links.remove(self)
        except ValueError as ee:
            self.clsLogger.warning(f'Ignoring destination removal error: {ee}')

    def optimizeLine(self):
        """
        Optimize line, so that the line length is minimized
        """
        # Get elements
        srcAnchor = self.GetSource()
        dstAnchor = self.GetDestination()

        srcX, srcY = self._srcShape.GetPosition()
        dstX, dstY = self._destShape.GetPosition()

        srcSize = self._srcShape.GetSize()
        dstSize = self._destShape.GetSize()

        OglLink.clsLogger.info(f"optimizeLine - ({srcX},{srcY}) / ({dstX},{dstY})")
        # Find new positions
        # Little tips
        optimalSrcX, optimalSrcY, optimalDstX, optimalDstY = dstX, dstY, srcX, srcY

        optimalSrcX += dstSize[0] // 2
        optimalSrcY += dstSize[1] // 2
        optimalDstX += srcSize[0] // 2
        optimalDstY += srcSize[1] // 2

        srcAnchor.SetPosition(optimalSrcX, optimalSrcY)
        dstAnchor.SetPosition(optimalDstX, optimalDstY)

    # noinspection PyUnusedLocal
    def OnRightDown(self, event: MouseEvent):
        """
        Handle right-clicks on our UML LineShape-  Override base handler;  It does nothing

        Args:
            event:
        """
        from ogl.sd.OglSDMessage import OglSDMessage

        if isinstance(self, OglSDMessage) is True:
            return
        menu: Menu = Menu()
        menu.Append(MENU_ADD_BEND,      'Add Bend',      'Add Bend at right click point')
        menu.Append(MENU_REMOVE_BEND,   'Remove Bend',   'Remove Bend closest to click point')
        menu.Append(MENU_TOGGLE_SPLINE, 'Toggle Spline', 'Best with at least one bend')

        if len(self._controls) == 0:
            bendItem: MenuItem = menu.FindItemById(MENU_REMOVE_BEND)
            bendItem.Enable(enable=False)

        x: int = event.GetX()
        y: int = event.GetY()
        clickPoint: Tuple[int, int] = (x, y)

        OglLink.clsLogger.debug(f'OglLink - {clickPoint=}')
        # I hate lambdas -- humberto
        menu.Bind(EVT_MENU, lambda evt, data=clickPoint: self._onMenuItemSelected(evt, data))

        frame = self._diagram.GetPanel()
        frame.PopupMenu(menu, x, y)

    # noinspection PyUnusedLocal
    def _onMenuItemSelected(self, event: CommandEvent, data):

        eventId: int = event.GetId()
        if eventId == MENU_ADD_BEND:
            self._addBend(data)
        elif eventId == MENU_REMOVE_BEND:
            self._removeBend(data)
        elif eventId == MENU_TOGGLE_SPLINE:
            self._toggleSpline()

    def _computeLinkLength(self, srcPosition: OglPosition, destPosition: OglPosition) -> int:
        """

        Returns:  The length of the link between the source shape and destination shape
        """
        dx, dy = self._computeDxDy(srcPosition, destPosition)
        linkLength = round(sqrt(dx*dx + dy*dy))
        if linkLength == 0:
            linkLength = 1

        return linkLength

    def _computeDxDy(self, srcPosition: OglPosition, destPosition: OglPosition) -> Tuple[int, int]:
        """

        Args:
            srcPosition:    source position
            destPosition:   destination position

        Returns:
            A tuple of deltaX and deltaY of the shape position
        """
        if self._srcShape is None or self._destShape is None:
            raise IllegalOperationException('Either the source or the destination shape is None')

        srcX: int = srcPosition.x
        srcY: int = srcPosition.y
        dstX: int = destPosition.x
        dstY: int = destPosition.y

        dx: int = dstX - srcX
        dy: int = dstY - srcY

        return dx, dy

    def _addBend(self, clickPoint: Tuple[int, int]):

        OglLink.clsLogger.debug(f'Add a bend.  {clickPoint=}')

        x = clickPoint[0]
        y = clickPoint[1]
        cp = ControlPoint(x, y)

        cp.SetVisible(True)
        #
        # Add it either before the destinationAnchor or the sourceAnchor
        #
        lp: LinePoint = self.GetSource()
        self.AddControl(control=cp, after=lp)

        frame = self._diagram.GetPanel()
        frame.GetDiagram().AddShape(cp)
        frame.Refresh()

    def _removeBend(self, clickPoint: Tuple[int, int]):

        OglLink.clsLogger.debug(f'Remove a bend.  {clickPoint=}')

        cp: ControlPoint = self._findClosestControlPoint(clickPoint=clickPoint)

        assert cp is not None, 'We should have previously verified there was at least one on the line'

        self._removeControl(control=cp)
        cp.Detach()
        cp.SetVisible(False)    # Work around still on screen but not visible and not saved

        frame = self._diagram.GetPanel()
        frame.Refresh()

    def _toggleSpline(self):

        self.SetSpline(not self.GetSpline())

        frame = self._diagram.GetPanel()
        frame.Refresh()

    def _findClosestControlPoint(self, clickPoint: Tuple[int, int]) -> ControlPoint:

        controlPoints = self.GetControlPoints()

        distance:     float        = 1000.0    # Impossibly long distance
        closestPoint: ControlPoint = cast(ControlPoint, None)
        srcPosition:  OglPosition  = OglPosition(x=clickPoint[0], y=clickPoint[1])

        for controlPoint in controlPoints:
            xy:    Tuple[int, int] = controlPoint.GetPosition()
            destX: int = xy[0]
            destY: int = xy[1]
            destPosition: OglPosition = OglPosition(x=destX, y=destY)

            dx, dy = self._computeDxDy(srcPosition, destPosition)
            currentDistance = sqrt(dx*dx + dy*dy)
            self.clsLogger.debug(f'{currentDistance=}')
            if currentDistance <= distance:
                distance = currentDistance
                closestPoint = cast(ControlPoint, controlPoint)

        return closestPoint

    def __repr__(self):

        srcShape: Shape = self.sourceShape
        dstShape: Shape = self.destinationShape
        sourceId: int   = srcShape.id
        dstId:    int   = dstShape.id

        return f'from: id: {sourceId} {srcShape} to id: {dstId} {dstShape}'




from typing import Tuple
from typing import cast

from wx import DC

from miniogl.Shape import Shape
from miniogl.MiniOglUtils import sign
from miniogl.SizerShape import SizerShape
from miniogl.RectangleShapeModel import RectangleShapeModel


class RectangleShape(Shape):
    """
    A rectangle shape.
    """
    def __init__(self, x: int = 0, y: int = 0, width: int = 0, height: int = 0, parent=None):
        """

        Args:
            x:      Shape's x position
            y:      Shape's y position
            width:  Shape's width
            height: Shape's height
            parent: The shape's parent, if any
        """
        super().__init__(x, y, parent)

        self._width:  int = width   # width and height can be < 0 !!!
        self._height: int = height

        self._drawFrame: bool = True
        self._resizable: bool = True

        self._topLeftSizer:  SizerShape = cast(SizerShape, None)
        self._topRightSizer: SizerShape = cast(SizerShape, None)
        self._botLeftSizer:  SizerShape = cast(SizerShape, None)
        self._botRightSizer: SizerShape = cast(SizerShape, None)

        self._ox: int = 0   # This is done in Shape but Pycharm can't see this in the ShowSizer() code
        # set the model of the shape (MVC pattern)
        self._model: RectangleShapeModel = RectangleShapeModel(self)
        self._model.SetPosition(x=x, y=y)
        self._model.SetSize(width=width, height=height)

    def SetResizable(self, state: bool):
        """
        Set the resizable flag.

        @param state
        """
        self._resizable = state

    def GetResizable(self) -> bool:
        """
        Get the resizable flag.

        @return bool
        """
        return self._resizable

    def GetTopLeft(self) -> Tuple[int, int]:
        """
        Get the coordinates of the top left point in diagram coordinates.

        @return (double, double)
        """
        x, y = self.GetPosition()
        x -= self._ox
        y -= self._oy
        width, height = self.GetSize()
        if width < 0:
            x += width
        if height < 0:
            y += height
        return x, y

    def SetTopLeft(self, x, y):
        """
        Set the position of the top left point.

        @param  x
        @param y : new position
        """
        x += self._ox
        y += self._oy
        width, height = self.GetSize()
        if width < 0:
            x -= width
        if height < 0:
            y -= height
        self._x, self._y = x, y

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Draw the rectangle on the dc.

        Args:
            dc:
            withChildren:

        Returns:

        """
        if self._visible:
            Shape.Draw(self, dc, False)
            if self._drawFrame:
                sx, sy = self.GetPosition()
                sx, sy = sx - self._ox, sy - self._oy
                width, height = self.GetSize()

                dc.DrawRectangle(sx, sy, width, height)
            if withChildren:
                self.DrawChildren(dc)
            if self._topLeftSizer is not None:
                self._topLeftSizer.Draw(dc, False)

    def DrawBorder(self, dc):
        """
        Draw the border of the shape, for fast rendering.
        """
        Shape.DrawBorder(self, dc)
        sx, sy = self.GetPosition()
        sx, sy = sx - self._ox, sy - self._oy
        width, height = self.GetSize()
        dc.DrawRectangle(sx, sy, width, height)

    def SetDrawFrame(self, draw):
        """
        Choose to draw a frame around the rectangle.

        @param bool draw : True to draw the frame
        """
        self._drawFrame = draw

    def Inside(self, x, y) -> bool:
        """
        True if (x, y) is inside the rectangle.

        @param  x
        @param y
        @return bool
        """
        # this also works if width and/or height is negative.
        sx, sy = self.GetPosition()
        # take a minimum of 4 pixels for the selection
        width, height = self.GetSize()
        width  = sign(width)  * max(abs(width),  4)
        height = sign(height) * max(abs(height), 4)
        topLeftX: int = sx - self._ox
        topLeftY: int = sy - self._oy

        topXLeftIsInside:  bool  = x > topLeftX
        topXRightIsInside: bool  = x < topLeftX + width
        topYLeftIsInside:  bool  = y > topLeftY
        topYRightIsInside: bool  = y < topLeftY + height
        if topXLeftIsInside and topXRightIsInside and topYLeftIsInside and topYRightIsInside:
            return True
        else:
            return False

    def GetSize(self) -> Tuple[int, int]:
        """
        Get the size of the rectangle.

        Returns:
            A tuple of width, height
        """
        return self._width, self._height

    def GetHeight(self):
        return self._height

    def SetSelected(self, state: bool = True):
        """
        Select the shape.

        @param state
        """
        Shape.SetSelected(self, state)
        if self._resizable:
            self.ShowSizers(state)

    def Detach(self):
        """
        Detach the shape from its diagram.
        This is the way to delete a shape. All anchor points are also
        removed, and link lines too.
        """
        # do not detach a protected shape
        if self._diagram is not None and not self._protected:
            Shape.Detach(self)
            self.ShowSizers(False)

    def ShowSizers(self, state: bool = True):
        """
        Show the four sizer shapes if state is True.

        @param state
        """
        width, height = self.GetSize()
        if state and not self._topLeftSizer:
            self._topLeftSizer  = SizerShape(-self._ox, -self._oy, self)
            self._topRightSizer = SizerShape(-self._ox + width - 1, self._oy, self)
            self._botLeftSizer  = SizerShape(-self._ox, -self._oy + height - 1, self)
            self._botRightSizer = SizerShape(-self._ox + width - 1, -self._oy + height - 1, self)

            self._diagram.AddShape(self._topLeftSizer)
            self._diagram.AddShape(self._topRightSizer)
            self._diagram.AddShape(self._botLeftSizer)
            self._diagram.AddShape(self._botRightSizer)
        elif not state and self._topLeftSizer is not None:
            self._topLeftSizer.Detach()
            self._topRightSizer.Detach()
            self._botLeftSizer.Detach()
            self._botRightSizer.Detach()

            self._topLeftSizer  = cast(SizerShape, None)
            self._topRightSizer = cast(SizerShape, None)
            self._botLeftSizer  = cast(SizerShape, None)
            self._botRightSizer = cast(SizerShape, None)

    def SetSize(self, width, height):
        """
        Set the size of the rectangle.

        @param  width
        @param height
        """
        self._width, self._height = width, height

        if self.HasDiagramFrame():
            self.UpdateModel()

        for anchor in self._anchors:
            ax, ay = anchor.GetPosition()
            # Reset position to stick the border
            anchor.SetPosition(ax, ay)

    def Resize(self, sizer, x, y):
        """
        Resize the rectangle according to the new position of the sizer.
        Not used to programmatically resize a shape. Use `SetSize` for this.

        Args:
            sizer:
            x:      x position of the sizer
            y:      y position of the sizer

        """
        tlx, tly = self.GetTopLeft()
        w, h = self.GetSize()
        sw, sh = sign(w), sign(h)
        w, h = abs(w), abs(h)
        if sizer is self._topLeftSizer:
            nw = sw * (w - x + tlx)
            nh = sh * (h - y + tly)
            self._ox = self._ox * nw // w
            self._oy = self._oy * nh // h
            self.SetSize(nw, nh)
            self.SetTopLeft(x, y)
            self._topRightSizer.SetRelativePosition(nw - 1, 0)
            self._botLeftSizer.SetRelativePosition(0, nh - 1)
            self._botRightSizer.SetRelativePosition(nw - 1, nh - 1)
        elif sizer is self._topRightSizer:
            nw = sw * (x - tlx)
            nh = sh * (tly + h - y)
            self.SetTopLeft(tlx, y)
            self.SetSize(nw + 1, nh)
            self._topRightSizer.SetRelativePosition(nw, 0)
            self._botRightSizer.SetRelativePosition(nw, nh - 1)
            self._botLeftSizer.SetRelativePosition(0, nh - 1)
        elif sizer is self._botLeftSizer:
            nw = sw * (w - x + tlx)
            nh = sh * (y - tly)
            self.SetTopLeft(x, tly)
            self.SetSize(nw, nh + 1)
            self._botLeftSizer.SetRelativePosition(0, nh)
            self._botRightSizer.SetRelativePosition(nw - 1, nh)
            self._topRightSizer.SetRelativePosition(nw - 1, 0)
        elif sizer is self._botRightSizer:
            nw = sw * (x - tlx)
            nh = sh * (y - tly)
            self.SetSize(nw + 1, nh + 1)
            self._botLeftSizer.SetRelativePosition(0, nh)
            self._botRightSizer.SetRelativePosition(nw, nh)
            self._topRightSizer.SetRelativePosition(nw, 0)

    def UpdateFromModel(self):
        """
        Updates the shape position and size from the model in the light of a
        change of state of the diagram frame (here it's only for the zoom)
        """

        #  change the position of the shape from the model
        Shape.UpdateFromModel(self)

        #  get the model size
        width, height = self.GetModel().GetSize()

        #  get the diagram frame ratio between the shape and the model
        ratio = self.GetDiagram().GetPanel().currentZoom

        # set the new size to the shape.
        self._width  = round(width * ratio)
        self._height = round(height * ratio)

    def UpdateModel(self):
        """
        Updates the model when the shape (view) is moved or resized.
        """
        #  change the coordinates of model
        Shape.UpdateModel(self)

        #  get the size of the shape (view)
        width, height = self.GetSize()

        # get the ratio between the model and the shape (view) from
        # the diagram frame where the shape is displayed.
        ratio = self.GetDiagram().GetPanel().currentZoom

        # set the new size to the model.
        self.GetModel().SetSize(round(width//ratio), round(height//ratio))




from miniogl.PointShape import PointShape


class SizerShape(PointShape):
    """
    A sizer, to resize other shapes.

    """
    def __init__(self, x, y, parent):
        """
        Constructor.

        @param double x, y : position of the point
        @param Shape parent : parent shape
        """
        #  print ">>>SizerShape", x, y
        PointShape.__init__(self, x, y, parent)
        self._moving = True

    def Draw(self, dc, withChildren=True):
        """
        #  TODO : Remove this. This is for debugging purpose.

        Note : This functions seems to be needed to display anchors
                on rectangle when moving them, but not for lines,
            single anchors, ...

        Args:
            dc:
            withChildren:

        Returns:

        """
        PointShape.Draw(self, dc, withChildren)
        pass

    def SetPosition(self, x, y):
        """
        Change the position of the shape, if it's draggable.

        @param  x
        @param y : new position
        """
        self._parent.Resize(self, x, y)
        # the position of the sizer is not changed, because it is relative
        # to the parent

    def SetMoving(self, state):
        """
        Set the moving flag.
        If setting a sizer moving, the parent will also be set moving.

        @param
        """
        PointShape.SetMoving(self, True)
        # a sizer is always moving
        self._parent.SetMoving(state)




from wx import DC

from ogl.OglObject import OglObject
from pyutmodel.PyutActor import PyutActor


MARGIN: int = 10


class OglActor(OglObject):
    """
    OGL object that represent a UML actor in use case diagrams.
    This class defines OGL objects that represents an actor for Use
    Cases diagram. You can just instantiate an OGL actor and add it to
    the diagram, links, resizing, ... are managed by parent class
    `OglObject`.

    For more instructions about how to create an OGL object, please refer
    to the `OglObject` class.

    :version: $Revision: 1.9 $
    :author: Philippe Waelti
    :contact: pwaelti@eivd.ch
    """
    def __init__(self, pyutActor=None, w: int = 80, h: int = 100):
        """

        Args:
            pyutActor:

            w:  width of shape
            h:  height of shape
        """

        # Init associated PyutObject
        if pyutActor is None:
            pyutObject = PyutActor()
        else:
            pyutObject = pyutActor

        super().__init__(pyutObject, w, h)
        self._drawFrame = False

    def Draw(self, dc: DC, withChildren: bool = False):
        """
        Draw an actor

        Args:
            dc:     The device context to draw on
            withChildren:   Draw the children or not

        """
        OglObject.Draw(self, dc)
        # Get current font
        dc.SetFont(self._defaultFont)

        # Gets the minimum bounding box for the shape
        width, height = self.GetSize()

        # Calculate the top center of the shape
        x, y = self.GetPosition()

        # drawing is restricted in the specified region of the device
        dc.SetClippingRegion(x, y, width, height)

        # Our sweet actor size
        actorWidth  = width
        actorHeight = int(0.8 * (height - 2.0 * MARGIN))  # 80 % of total height
        sizer = min(actorHeight, actorWidth)

        # Draw our actor head
        centerX = x + width  // 2
        centerY = y + height // 2

        x = int(centerX - 0.2 * sizer)
        y += MARGIN
        percentageSizer: int = int(0.4 * sizer)
        # dc.DrawEllipse(x, y, 0.4 * sizer, 0.4 * sizer)
        dc.DrawEllipse(x, y, percentageSizer, percentageSizer)

        # Draw body and arms
        x = centerX
        y += round(0.4 * sizer)
        # dc.DrawLine(x, y, x, y + 0.3 * actorHeight)
        # dc.DrawLine(x - 0.25 * actorWidth, y + 0.15 * actorHeight,
        #             x + 0.25 * actorWidth, y + 0.15 * actorHeight)
        dc.DrawLine(x, y, x, y + round(0.3 * actorHeight))
        dc.DrawLine(round(x - 0.25 * actorWidth), round(y + 0.15 * actorHeight),
                    round(x + 0.25 * actorWidth), round(y + 0.15 * actorHeight))

        # And the feet
        # y += round(0.3 * actorHeight)
        # dc.DrawLine(x, y, x - 0.25 * actorWidth, y + 0.3 * actorHeight)
        # dc.DrawLine(x, y, x + 0.25 * actorWidth, y + 0.3 * actorHeight)

        actorFeetPercentage: int = round(0.3 * actorHeight)
        y += round(actorFeetPercentage)
        # dc.DrawLine(x, y, x - 0.25 * actorWidth, y + actorFeetPercentage)
        # dc.DrawLine(x, y, x + 0.25 * actorWidth, y + actorFeetPercentage)
        dc.DrawLine(x, y, x - round(0.25 * actorWidth), y + actorFeetPercentage)
        dc.DrawLine(x, y, x + round(0.25 * actorWidth), y + actorFeetPercentage)

        # Draw our buddy name
        textWidth, textHeight = dc.GetTextExtent(self.pyutObject.name)

        # y = centerY + 0.5 * height - MARGIN - 0.1 * actorHeight
        y = round(centerY + 0.5 * height - MARGIN - 0.1 * actorHeight)

        # dc.DrawText(self.getPyutObject().getName(), x - 0.5 * textWidth, y)
        dc.DrawText(self.pyutObject.name, round(x - 0.5 * textWidth), y)
        dc.DestroyClippingRegion()




from typing import Tuple

from logging import Logger
from logging import getLogger

from wx import DC
from wx import FONTSTYLE_NORMAL
from wx import FONTWEIGHT_NORMAL
from wx import Font

from miniogl.Common import Common
from miniogl.AttachmentLocation import AttachmentLocation
from miniogl.SelectAnchorPoint import SelectAnchorPoint
from miniogl.LollipopLine import LollipopLine

from pyutmodel.PyutInterface import PyutInterface
from pyutmodel.PyutObject import PyutObject

from ogl.OglPosition import OglPosition
from ogl.OglTextFontFamily import OglTextFontFamily

from ogl.preferences.OglPreferences import OglPreferences

from ogl.OglUtils import OglUtils


class OglInterface2(LollipopLine, Common):

    ADJUST_AWAY_FROM_IMPLEMENTOR: int = 10

    INTERFACE_FONT_SIZE: int = 12   # TODO:  Make this a preference

    def __init__(self, pyutInterface: PyutInterface,  destinationAnchor: SelectAnchorPoint):

        LollipopLine.__init__(self, destinationAnchor=destinationAnchor)

        self.logger: Logger = getLogger(__name__)

        self._pyutInterface: PyutInterface = pyutInterface

        preferences: OglPreferences = OglPreferences()

        fontStyle:  OglTextFontFamily = preferences.textFontFamily
        fontFamily: int              = OglUtils.oglFontFamilyToWxFontFamily(fontStyle)

        self._defaultFont: Font = Font(OglInterface2.INTERFACE_FONT_SIZE, fontFamily, FONTSTYLE_NORMAL, FONTWEIGHT_NORMAL)

    @property
    def pyutInterface(self) -> PyutInterface:
        return self._pyutInterface

    @pyutInterface.setter
    def pyutInterface(self, theNewValue: PyutInterface):
        self._pyutInterface = theNewValue

    @property
    def pyutObject(self) -> PyutObject:
        """
        Synonym for reverse compatibility

        Returns:  The pyut interface
        """
        return self.pyutInterface

    @pyutObject.setter
    def pyutObject(self, newValue: PyutInterface):
        """
        Synonym for reverse compatibility

        Args:
            newValue: The new PyutInterface object
        """
        self._pyutInterface = newValue

    def Draw(self, dc: DC, withChildren: bool = True):

        super().Draw(dc=dc, withChildren=withChildren)
        dc.SetFont(self._defaultFont)

        xFaceName: str = self.pyutInterface.name

        extentSize: Tuple[int, int] = dc.GetTextExtent(xFaceName)  # width, height

        pixelSize: Tuple[int, int] = self._defaultFont.GetPixelSize()

        textPosition: OglPosition = self._determineInterfaceNamePosition(self._destinationAnchor, pixelSize=pixelSize, textSize=extentSize)

        dc.DrawText(xFaceName, textPosition.x, textPosition.y)

    def Inside(self, clickPointX, clickPointY) -> bool:
        """
        Override Shape.Inside

        Args:
            clickPointX: x click coordinate
            clickPointY: y click coordinate

        Returns:  `True` if (x, y) is inside the shape.
        """
        clickDiffStartX, clickDiffStartY, diffX, diffY = self.setupInsideCheck(clickPointX=clickPointX, clickPointY=clickPointY, line=self.lineCoordinates())

        if self.insideBoundingBox(clickDiffStartX, clickDiffStartY, diffX, diffY) and self.insideSegment(clickDiffStartX, clickDiffStartY, diffX, diffY):
            return True

        return False

    def UpdateFromModel(self):
        super().UpdateFromModel()

    def Remove(self, point):
        """
        Need an empty implementation of this to allow adding of additional interfaces
        on an OglClass;  This confirms I really do not understand the MiniOgl structure

        Args:
            point:
        """
        self.logger.debug(f'{point=}')

    def __repr__(self):

        strMe: str = f'OglInterface2 - "{self._pyutInterface.name}"'
        return strMe

    def __eq__(self, other):

        if isinstance(other, OglInterface2):
            if self._isSameName(other) is True and self._isSameId(other) is True:
                return True
            else:
                return False
        else:
            return False

    def __hash__(self):
        return hash(self._pyutInterface.name) + hash(self.id)

    def _isSameName(self, other) -> bool:

        ans: bool = False
        if self.pyutInterface.name == other.pyutInterface.name:
            ans = True
        return ans

    def _isSameId(self, other):

        ans: bool = False
        # if self.GetID() == other.GetID():
        if self.id == other.id:
            ans = True
        return ans

    def _determineInterfaceNamePosition(self, destinationAnchor: SelectAnchorPoint, pixelSize: Tuple[int, int], textSize: Tuple[int, int]) -> OglPosition:

        oglPosition:     OglPosition     = OglPosition()
        attachmentPoint: AttachmentLocation = destinationAnchor.attachmentPoint

        x, y = destinationAnchor.GetPosition()

        fWidth, fHeight = pixelSize
        tWidth, tHeight = textSize

        if attachmentPoint == AttachmentLocation.NORTH:
            y -= (LollipopLine.LOLLIPOP_LINE_LENGTH + (LollipopLine.LOLLIPOP_CIRCLE_RADIUS * 2) + OglInterface2.ADJUST_AWAY_FROM_IMPLEMENTOR)
            x -= (tWidth // 2)
            oglPosition.x = x
            oglPosition.y = y

        elif attachmentPoint == AttachmentLocation.SOUTH:
            y += (LollipopLine.LOLLIPOP_LINE_LENGTH + LollipopLine.LOLLIPOP_CIRCLE_RADIUS + OglInterface2.ADJUST_AWAY_FROM_IMPLEMENTOR)
            x -= (tWidth // 2)
            oglPosition.x = x
            oglPosition.y = y

        elif attachmentPoint == AttachmentLocation.WEST:
            y = y - (fHeight * 2)
            originalX: int = x
            x = x - LollipopLine.LOLLIPOP_LINE_LENGTH - (tWidth // 2)
            while x + tWidth > originalX:
                x -= OglInterface2.ADJUST_AWAY_FROM_IMPLEMENTOR
            oglPosition.x = x
            oglPosition.y = y

        elif attachmentPoint == AttachmentLocation.EAST:
            y = y - (fHeight * 2)
            x = x + round(LollipopLine.LOLLIPOP_LINE_LENGTH * 0.8)
            oglPosition.x = x
            oglPosition.y = y
        else:
            self.logger.warning(f'Unknown attachment point: {attachmentPoint}')
            assert False, 'Unknown attachment point'

        return oglPosition



from typing import cast

from logging import Logger
from logging import getLogger

from wx import BLACK_PEN
from wx import RED_PEN
from wx import DC

from miniogl.AttachmentLocation import AttachmentLocation
from miniogl.Common import CommonLine
from miniogl.Common import CommonPoint

from miniogl.SelectAnchorPoint import SelectAnchorPoint
from miniogl.Shape import Shape


class LollipopLine(Shape):

    LOLLIPOP_LINE_LENGTH:   int = 60
    LOLLIPOP_CIRCLE_RADIUS: int = 4

    def __init__(self, destinationAnchor: SelectAnchorPoint):

        super().__init__()

        self.logger:             Logger      = getLogger(__name__)
        self._destinationAnchor: SelectAnchorPoint = cast(SelectAnchorPoint, None)

        if destinationAnchor is not None:
            self._destinationAnchor = destinationAnchor
            destinationAnchor.AddLine(self)

    @property
    def destinationAnchor(self) -> SelectAnchorPoint:
        return self._destinationAnchor

    @destinationAnchor.setter
    def destinationAnchor(self, theNewValue: SelectAnchorPoint):
        self._destinationAnchor = theNewValue

    def lineCoordinates(self) -> CommonLine:

        attachmentPoint: AttachmentLocation = self._destinationAnchor.attachmentPoint

        xDest, yDest = self._destinationAnchor.GetPosition()
        circleX, circleY, xSrc, ySrc = self._calculateWhereToDrawLollipop(attachmentPoint, xDest, yDest)

        return CommonLine(CommonPoint(xSrc, ySrc), CommonPoint(xDest, yDest))

    def Draw(self, dc: DC, withChildren: bool = True):

        if self._selected:
            dc.SetPen(RED_PEN)
        else:
            dc.SetPen(BLACK_PEN)

        xDest, yDest = self._destinationAnchor.GetPosition()
        attachmentPoint: AttachmentLocation = self._destinationAnchor.attachmentPoint

        circleX, circleY, xSrc, ySrc = self._calculateWhereToDrawLollipop(attachmentPoint, xDest, yDest)

        self.logger.debug(f'Source: ({xSrc},{ySrc}) - Dest ({xDest},{yDest})')
        dc.DrawLine(xSrc, ySrc, xDest, yDest)
        dc.DrawCircle(circleX, circleY, LollipopLine.LOLLIPOP_CIRCLE_RADIUS)

    def _calculateWhereToDrawLollipop(self, attachmentPoint, xDest, yDest):
        """

        Args:
            attachmentPoint:
            xDest:
            yDest:

        Returns:  A tuple that is the x,y position of the circle and the end
        of the line
        """

        diagram = self.GetDiagram()
        panel   = diagram.GetPanel()   # to enable debugging and unit tests
        ratio = panel.currentZoom

        lollipopLength: int = LollipopLine.LOLLIPOP_LINE_LENGTH * ratio
        self.logger.debug(f'{lollipopLength}')

        if attachmentPoint == AttachmentLocation.EAST:
            xSrc:    int = int(xDest + lollipopLength)
            ySrc:    int = int(yDest)
            circleX: int = int(xDest + lollipopLength)
            circleY: int = int(yDest)
        elif attachmentPoint == AttachmentLocation.WEST:
            xSrc    = int(xDest - lollipopLength)
            ySrc    = int(yDest)
            circleX = int(xDest - lollipopLength)
            circleY = int(yDest)
        elif attachmentPoint == AttachmentLocation.NORTH:
            xSrc    = int(xDest)
            ySrc    = int(yDest - lollipopLength)
            circleX = int(xDest)
            circleY = int(yDest - lollipopLength)
        else:  # it is South
            xSrc    = int(xDest)
            ySrc    = int(yDest + lollipopLength)
            circleX = int(xDest)
            circleY = int(yDest + lollipopLength)

        return circleX, circleY, xSrc, ySrc
